/**
 * Constants
 */
var gc = gc || {};

var numPending = 0;
var maxNumPending = 0;

var UART_9600_bps = 0;
var UART_19200_bps = 1;
var UART_38400_bps = 2;
var UART_57600_bps = 3;
var UART_115200_bps = 4;
var UART_230400_bps = 5;
var UART_300_bps = 6;
var UART_320_bps = 7;
var UART_600_bps = 8;
var UART_1200_bps = 9;
var UART_2400_bps = 10;
var UART_4800_bps = 11;
var UART_460800_bps = 12;
var UART_921600_bps = 13;
var UART_Disable = 99;

var DEVICE_ADDR = 0x1;

var Cmd_UART_Control = 15; // 0x0F
var Cmd_UART_SetMode = 19; // 0x13

var Cmd_UART_Write = 18; // 0x12
var Cmd_UART_SetMode = 19; // 0x13
var Cmd_UART_Read = 20; // 0x14

var COMMAND_PACKET = 1;
var REPLY_PACKET = 2;
var ERROR_PACKET = 3;
var PAYLOAD_PACKET = 4;
var INTERRUPT_PACKET = 5;

// packet flags
var FLAG_MOREDATA = 1;

var PACKET_IDENTIFIER = 'T'.charCodeAt(0);

var PACKET_ID = 0;
var PACKET_PEC = 1;
var PACKET_PAYLOAD_LEN = 2;
var PACKET_TYPE = 3;
var PACKET_FLAGS = 4;
var PACKET_SEQ_NUM = 5;
var PACKET_STATUS = 6;
var PACKET_COMMAND = 7;
var PACKET_PAYLOAD = 8;

var MAX_PACKET_SIZE = 62; // size, in bytes, of a USB packet
var PACKET_HEADER_SIZE = PACKET_PAYLOAD;
var MAX_PAYLOAD = MAX_PACKET_SIZE - PACKET_HEADER_SIZE;

/**
 * Helper functions
 */

var parseStringProperty = function(message, value, valueMap) {
  var stringValue = ('' + value).toLowerCase();
  if (valueMap.hasOwnProperty(stringValue)) {
    return valueMap[stringValue];
  } else {
    message =
      TheValue +
      value +
      ForEntry +
      message +
      InRegDefs +
      ' is not supported.  Valid entries are';
    var delimiter = ' "';
    var lastOption;
    for (var option in valueMap) {
      if (valueMap.hasOwnProperty(option)) {
        if (lastOption) {
          message = message + delimiter + lastOption;
          delimiter = '", "';
        }
        lastOption = option;
      }
    }
    throw message + '", or "' + lastOption + '".';
  }
};

var TheValue = 'The value "';
var ForEntry = '", for entry ';
var InRegDefs = ', in the register-defs.json file';
var parseNumberProperty = function(message, value, min, max) {
  var result = +value;
  if (isNaN(result)) {
    throw TheValue +
      value +
      ForEntry +
      message +
      InRegDefs +
      ' is not a number.';
  }
  if (min && result < min) {
    throw TheValue +
      value +
      ForEntry +
      message +
      InRegDefs +
      ' must be greater than ' +
      min +
      '.';
  }
  if (max && result > max) {
    throw TheValue +
      value +
      ForEntry +
      message +
      InRegDefs +
      ' must be less than ' +
      max +
      '.';
  }
  return result;
};

var getDeviceIndexUart = function(coreIndex, regModelId) {
  coreIndex = coreIndex === undefined ? 0 : coreIndex;

  if (
    coreIndex > -1 &&
    regModelId === 'bq7971x' &&
    gc.databind.registry.getModel('bq79600')._codec.getInterfaces()[
      'customuart'
    ].is600
  ) {
    coreIndex = coreIndex + 1;
  }

  return coreIndex;
};

var init = function() {
  var regReadDetail = undefined,
    regReadCount,
    regReadPromiseArr;

  /**
   * This code implements an example custom USB2ANY Interface (functionally just a clone of the default I2C interface)
   *
   * @constructor
   * @param settings - The interface settings defined in the system.json
   * @param registerModel - The register model
   * @implements gc.databind.USB2ANY.ICommInterface
   */
  CustomUART_Interface = function(settings, registerModel) {
    this._registerModel = registerModel;
    this.transferPromises = [];
    this.cmdQueue = [];
    this.readyToProcess = true;
    this.noOfDevices = 0;
    this.config = {};
    this.is600 = false;
  };

  /**
   * Add the custom interface to the USB2ANY prototype, the name must take the form of "XXX_Interface" where "XXX" is all capitalized, eg. "LED_Interface"
   * To use custom interfaces in the system.json, set the name property to be the "XXX" part of the interface name (not case-sensitive), eg. "name" : "led"
   * To avoid naming conflicts with already defined interfaces, do not use names "power", "i2c", "spi", "gpio", "uart", "pwm" or "displayscale"
   */
  gc.databind.USB2ANY.prototype.CUSTOMUART_Interface = CustomUART_Interface;

  /**
   * Custom interfaces should be extended from the ICommInterface base class
   */
  CustomUART_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();

  CustomUART_Interface.prototype._parityMap = {none: 0, even: 1, odd: 2};
  CustomUART_Interface.prototype._bitDirectionMap = {lsb: 0, msb: 1};

  /**
   * Initialize symbols for device
   *
   * @param settings - The interface settings defined in the system.json
   * @param registerModel - The register model
   */
  /*      CustomUART_Interface.prototype.initSymbolsForDevice = function(settings, registerModel) {
         //registerModel.readDeviceAddrsMap('UART', settings);
         var c = gc.databind.PacketCodecFactory.getConstructor(settings.codec);
         if (c && c.prototype && c.prototype.initSymbolsForDevice)
         {
             c.prototype.initSymbolsForDevice(settings, registerModel);
         }
     }; */

  /**
   * Set the communication parameters
   *
   * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
   * @param settings - The interface settings defined in the system.json
   */
  CustomUART_Interface.prototype.control = function(u2a, settings) {
    console.log('UART', 'Control', u2a, settings);

    var baudRate = UART_2400_bps;
    switch (+settings.baudRate) {
      case 19200:
        baudRate = UART_19200_bps;
        break;
      case 38400:
        baudRate = UART_38400_bps;
        break;
      case 57600:
        baudRate = UART_57600_bps;
        break;
      case 115200:
        baudRate = UART_115200_bps;
        break;
      case 230400:
        baudRate = UART_230400_bps;
        break;
      case 300:
        baudRate = UART_300_bps;
        break;
      case 320:
        baudRate = UART_320_bps;
        break;
      case 600:
        baudRate = UART_600_bps;
        break;
      case 1200:
        baudRate = UART_1200_bps;
        break;
      case 2400:
        baudRate = UART_2400_bps;
        break;
      case 4800:
        baudRate = UART_4800_bps;
        break;
      case 460800:
        baudRate = UART_460800_bps;
        break;
      case 921600:
        baudRate = UART_921600_bps;
        break;
    }

    var parity = parseStringProperty(
      'parity',
      settings.parity,
      this._parityMap
    );
    var bitDirection = parseStringProperty(
      'bitDirection',
      settings.bitDirection,
      this._bitDirectionMap
    );
    var characterLength =
      8 -
      parseNumberProperty(
        'characterLength',
        settings.characterLength || 8,
        7,
        8
      );
    var stopBits =
      parseNumberProperty('stopBits', settings.stopBits || 1, 1, 2) === 1
        ? 0
        : 1;

    this.config.baudRate = baudRate;
    this.config.parity = parity;
    this.config.bitDirection = bitDirection;
    this.config.characterLength = characterLength;
    this.config.stopBits = stopBits;
  };

  CustomUART_Interface.prototype.configure = function(u2a) {
    this.enable(u2a);

    u2a.onPayloadPacketReceived = this.onPayloadPacketReceived;
    u2a.shouldPauseTransmission = this.shouldPauseTransmission;

    for (var i = 0; i < this.transferPromises.length; i++) {
      this.transferPromises[i].reject();
    }
    this.cmdQueue.splice(0, this.cmdQueue.length);
    this.readyToProcess = true;
  };

  CustomUART_Interface.prototype.enable = function(u2a) {
    console.log('UART', 'Config', 'Params', [
      this.config.baudRate,
      this.config.parity,
      this.config.bitDirection,
      this.config.characterLength,
      this.config.stopBits
    ]);
    u2a.sendCommandPacket(Cmd_UART_Control, [
      this.config.baudRate,
      this.config.parity,
      this.config.bitDirection,
      this.config.characterLength,
      this.config.stopBits
    ]);

    u2a.sendCommandPacket(Cmd_UART_SetMode, [0]);
  };

  CustomUART_Interface.prototype.disable = function(u2a) {
    u2a.sendCommandPacket(15, [99, 0, 0, 0, 0]);
  };

  CustomUART_Interface.prototype.calculateWriteData = function(writeData) {
    var bytesToRead = 0,
      bytesToReadHigh,
      bytesToReadLow,
      waitTime = 0,
      waitTimeHigh,
      waitTimeLow,
      devAddr;
    var initByte = writeData[0];
    var noOfDevices = this.noOfDevices;

    if (initByte === 0x80) {
      devAddr = writeData[1];
      bytesToRead = writeData[4] + 7;
      // if(writeData[1] === 0x00) {
      //     waitTime = 100 + (bytesToRead * 10);
      // } else {
      //     waitTime = ((devAddr - 1) * 6) + (bytesToRead * 10) + 100;
      // }
    } else if (initByte === 0xa0) {
      bytesToRead =
        (writeData[3] + 7) * (noOfDevices > 0 ? noOfDevices - 1 : 0);
      // waitTime = ((noOfDevices > 2 ? noOfDevices - 2 : 0) * 6) + (bytesToRead * 10) + 100;
    } else if (initByte === 0xc0) {
      bytesToRead = (writeData[3] + 7) * noOfDevices;
      // waitTime = ((noOfDevices - 1) * 6) + (bytesToRead * 10) + 100;
    }

    bytesToReadHigh = Math.floor(bytesToRead / 256);
    bytesToReadLow = bytesToRead % 256;

    waitTime = 30000; //Wait time in micro seconds
    waitTimeHigh = Math.floor(waitTime / 256);
    waitTimeLow = waitTime % 256;

    writeData.unshift(writeData.length);
    writeData.push(bytesToReadHigh);
    writeData.push(bytesToReadLow);
    writeData.push(waitTimeHigh);
    writeData.push(waitTimeLow);

    return writeData;
  };

  CustomUART_Interface.prototype.shouldPauseTransmission = function(txPacket) {
    return false;
  };

  CustomUART_Interface.prototype.onPayloadPacketReceived = function(
    target,
    cmd,
    rawDataSliced,
    rawData,
    payloadLen,
    payloadStart
  ) {
    // console.log('UART', 'onPayloadPacketReceived', target, cmd, rawData);
    if (cmd === 144 || cmd === 146) {
      var transferPromises = this.getInterfaces()['customuart']
        .transferPromises;
      // console.log('UART', 'Transfer data in buffer', transferPromises.length);
      if (transferPromises && transferPromises.length > 0) {
        // console.log('UART', 'Transfer promise resolve');
        var transferPromise = transferPromises.shift();
        // console.log('UART', 'Transfer promise shifted', transferPromises.length);
        transferPromise.resolve({
          rawData: rawData,
          payloadLen: payloadLen,
          payloadStart: payloadStart
        });
      }
    } else if (cmd === 149) {
      console.log(
        'UART',
        'onPayloadPacketReceived',
        target,
        cmd,
        rawData,
        payloadLen
      );
      var status = rawData.slice(
        payloadStart,
        payloadStart + rawData[payloadLen]
      )[0];
      var uartInterface = this.getInterfaces()['customuart'];
      if (status > 0) {
        console.log('UART', 'onPayloadPacketReceived', 'Timeout');
        // uartInterface.disable(this);
        // this.sendCommandPacket(88, [4, 0, 0, 20]);
        // uartInterface.enable(this);
      }
      if (uartInterface.opPromise) {
        console.log('UART', 'onPayloadPacketReceived', 'Resolve');
        uartInterface.opPromise.resolve();
        uartInterface.opPromise = undefined;
      }
    } else {
      gc.console.debug('USB2ANY', 'Ignoring Payload Packet for command #', cmd);
    }
  };

  CustomUART_Interface.prototype.recursivePing = function(u2a, count) {
    var deferred = Q.defer();

    // Clear Rx Buffer
    u2a.sendCommandPacket(147, []);

    // 606
    // var writeData = [0x80, count, 0x01, 0x04, 0x00];

    // 616
    var writeData = [0x80, count, 0x03, 0x06, 0x00];

    var crc = getByteArrFromHexStr(
      prependZero(getCRC16IBM(writeData).toString(16), 4)
    );
    writeData.push.apply(writeData, crc);
    writeData = this.calculateWriteData(writeData);

    var opPromise = Q.defer();
    this.opPromise = opPromise;

    // setTimeout(function() {
    opPromise.promise.then(
      function() {
        var buffLenDeferred = Q.defer();
        this.transferPromises.push(buffLenDeferred);
        buffLenDeferred.promise.then(
          function(data) {
            var rawData = data.rawData;
            var payloadLen = data.payloadLen;
            var payloadStart = data.payloadStart;
            var payload = rawData.slice(
              payloadStart,
              payloadStart + rawData[payloadLen]
            );
            var dataLen = payload[1] * 256 + payload[0];

            if (dataLen > 0) {
              this.recursivePing(u2a, count + 1).then(
                function(countData) {
                  deferred.resolve(countData);
                }.bind(this)
              ).fail(function (err) {
                deferred.promise.reject(err);
              });
            } else {
              deferred.resolve(count);
            }
          }.bind(this)
        ).fail(function (err) {
          deferred.promise.reject(err);
        });
        u2a.sendCommandPacket(146, []);
        // }.bind(this), getReadCmdTimeout(1, 1));
      }.bind(this)
    );

    u2a.sendCommandPacket(149, writeData);

    return deferred.promise;
  };

  CustomUART_Interface.prototype.readFromU2A = function(u2a, readData) {
    var readDeferred = Q.defer();

    var bufferLenDeferred = Q.defer();
    this.transferPromises.push(bufferLenDeferred);
    bufferLenDeferred.promise.then(
      function(data) {
        // console.log('UART', 'Recursive read', 'Buffer data length', data);
        var rawData = data.rawData;
        var payloadLen = data.payloadLen;
        var payloadStart = data.payloadStart;
        var payload = rawData.slice(
          payloadStart,
          payloadStart + rawData[payloadLen]
        );

        var dataLen = payload[1] * 256 + payload[0];
        var packetSize = Math.ceil(dataLen / 54);

        function packetsTransferResolve(packet) {
          // console.log('UART', 'Recursive read', 'Packet', packet);
          var packetPayloadLen = packet.rawData[packet.payloadLen];
          var packetPayload = packet.rawData.slice(
            packet.payloadStart,
            packet.payloadStart + packetPayloadLen
          );
          readData.push.apply(readData, packetPayload);
          // console.log('UART', 'Recursive read', 'Res data', readData);
        }

        var packetDeferred;
        var packetsTransferPromise = [];
        for (var i = 0; i < packetSize; i++) {
          packetDeferred = Q.defer();
          packetsTransferPromise[i] = packetDeferred.promise;
          this.transferPromises.push(packetDeferred);
          packetsTransferPromise[i].then(packetsTransferResolve);
        }
        // console.log('UART', 'Recursive read', 'Packet transfer promises', packetsTransferPromise);
        Promise.all(packetsTransferPromise).then(
          function() {
            console.log('UART', 'Recursive read', 'Final res', readData);

            readDeferred.resolve(readData);
          }.bind(this)
        ).fail(function (err) {
          readDeferred.promise.reject(err);
        });

        if (packetSize > 0) {
          // console.log('UART', 'Recursive read', 'Buffer data transfer', (new Date()).getTime());
          u2a.sendCommandPacket(144, [payload[1], payload[0]]);
        }
      }.bind(this)
    );
    // console.log('UART', 'Recursive read', 'Get buffer size', 'Timestamp', (new Date()).getTime());
    u2a.sendCommandPacket(146, []);

    return readDeferred.promise;
  };

  CustomUART_Interface.prototype.recursiveRead = function(
    u2a,
    cmdArr,
    deferred,
    readData,
    calculatedBufferSize
  ) {
    var cmd = cmdArr.shift();
    logger('Read', cmd);

    // console.log('UART', 'Recursive read', 'Calculated read length', getReadDataLength(cmd, this.noOfDevices));

    if (getReadDataLength(cmd, this.noOfDevices) > 4824) {
      this.readFromU2A(u2a, readData).then(
        function(readData) {
          var tempReadData = [];

          var startReg = cmd[1] * 256 + cmd[2];
          var regCount = cmd[3] + 1;

          var firstStartReg = startReg;
          var firstRegCount = Math.floor(regCount / 2);

          var secondStartReg = startReg + firstRegCount;
          var secondRegCount = regCount - firstRegCount;

          var firstCmd = [cmd[0], cmd[1], cmd[2], firstRegCount - 1];
          var crc = getByteArrFromHexStr(
            prependZero(getCRC16IBM(firstCmd).toString(16), 4)
          );
          firstCmd.push.apply(firstCmd, crc);
          firstCmd = this.calculateWriteData(firstCmd);

          var secondCmd = [
            cmd[0],
            Math.floor(secondStartReg / 256),
            secondStartReg % 256,
            secondRegCount - 1
          ];
          crc = getByteArrFromHexStr(
            prependZero(getCRC16IBM(secondCmd).toString(16), 4)
          );
          secondCmd.push.apply(secondCmd, crc);
          secondCmd = this.calculateWriteData(secondCmd);

          var opPromiseFirst = Q.defer();
          this.opPromise = opPromiseFirst;

          // setTimeout(function() {
          opPromiseFirst.promise.then(
            function() {
              this.readFromU2A(u2a, tempReadData).then(
                function(tempReadData) {
                  var opPromiseSecond = Q.defer();
                  this.opPromise = opPromiseSecond;

                  // setTimeout(function() {
                  opPromiseSecond.promise.then(
                    function() {
                      this.readFromU2A(u2a, tempReadData).then(
                        function(tempReadData) {
                          tempReadData = parseDeviceResponse(tempReadData);

                          for (var i = 0; i < tempReadData.length; i++) {
                            for (var j = i + 1; j < tempReadData.length; j++) {
                              if (tempReadData[i][1] === tempReadData[j][1]) {
                                tempReadData[i][0] =
                                  tempReadData[i][0] + tempReadData[j][0] + 1;
                                tempReadData[i].splice(
                                  tempReadData[i].length - 2,
                                  2
                                );
                                tempReadData[j].splice(
                                  tempReadData[i].length - 2,
                                  2
                                );

                                for (
                                  var k = 4;
                                  k < tempReadData[j].length;
                                  k++
                                ) {
                                  tempReadData[i].push(tempReadData[j][k]);
                                }

                                crc = getByteArrFromHexStr(
                                  prependZero(
                                    getCRC16IBM(tempReadData[i]).toString(16),
                                    4
                                  )
                                );
                                tempReadData[i].push.apply(
                                  tempReadData[i],
                                  crc
                                );

                                tempReadData.splice(j, 1);
                                j--;
                              }
                            }
                          }

                          for (var i = 0; i < tempReadData.length; i++) {
                            readData.push.apply(readData, tempReadData[i]);
                          }

                          // console.log('UART', 'Recursive read', 'Appended data', readData);

                          if (cmdArr.length > 0) {
                            this.recursiveRead(
                              u2a,
                              cmdArr,
                              deferred,
                              readData,
                              0
                            );
                          } else {
                            logger('Response', readData);
                            deferred.resolve(readData);
                          }
                        }.bind(this)
                      ).fail(function (err) {
                        deferred.promise.reject(err);
                      });
                    }.bind(this)
                  );

                  u2a.sendCommandPacket(149, secondCmd);
                }.bind(this)
              );
            }.bind(this)
          );

          u2a.sendCommandPacket(149, firstCmd);
        }.bind(this)
      ).fail(function (err) {
        deferred.promise.reject(err);
      });
    } else {
      var noOfRegs = (cmd[0] < 160 ? cmd[4] : cmd[3]) + 1;
      var noOfDevices = cmd[0] < 160 ? 1 : this.noOfDevices;

      calculatedBufferSize =
        calculatedBufferSize + getReadDataLength(cmd, this.noOfDevices);

      cmd = this.calculateWriteData(cmd);

      if (calculatedBufferSize > 4824) {
        this.readFromU2A(u2a, readData).then(
          function(readData) {
            calculatedBufferSize = 0;

            // console.log('UART', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

            var opPromise = Q.defer();
            this.opPromise = opPromise;

            // setTimeout(function() {
            opPromise.promise.then(
              function() {
                if (cmdArr.length > 0) {
                  this.recursiveRead(
                    u2a,
                    cmdArr,
                    deferred,
                    readData,
                    calculatedBufferSize
                  );
                } else {
                  this.readFromU2A(u2a, readData).then(
                    function(readData) {
                      logger('Response', readData);
                      deferred.resolve(readData);
                    }.bind(this)
                  ).fail(function (err) {
                    deferred.promise.reject(err);
                  });
                }
                // }.bind(this), getReadCmdTimeout(noOfRegs, noOfDevices));
              }.bind(this)
            );

            u2a.sendCommandPacket(149, cmd);
          }.bind(this)
        );
      } else {
        // console.log('UART', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

        var opPromise = Q.defer();
        this.opPromise = opPromise;

        // setTimeout(function() {
        opPromise.promise.then(
          function() {
            if (cmdArr.length > 0) {
              this.recursiveRead(
                u2a,
                cmdArr,
                deferred,
                readData,
                calculatedBufferSize
              );
            } else {
              this.readFromU2A(u2a, readData).then(
                function(readData) {
                  logger('Response', readData);
                  deferred.resolve(readData);
                }.bind(this)
              ).fail(function (err) {
                deferred.promise.reject(err);
              });
            }
            // }.bind(this), getReadCmdTimeout(noOfRegs, noOfDevices));
          }.bind(this)
        );

        u2a.sendCommandPacket(149, cmd);
      }
    }
  };

  CustomUART_Interface.prototype.manageReads = function(u2a, cmdArr) {
    var deferred = Q.defer();
    var readData = [];

    // Clear Rx Buffer
    u2a.sendCommandPacket(147, []);

    if (cmdArr.length > 0) {
      this.recursiveRead(u2a, cmdArr, deferred, readData, 0);
    } else {
      deferred.resolve(readData);
    }
    return deferred.promise;
  };

  CustomUART_Interface.prototype.recursiveWrite = function(
    u2a,
    cmdArr,
    deferred
  ) {
    var cmd = cmdArr.shift();
    logger('Write', cmd);
    // console.log('UART', 'Recursive write', 'Timestamp', (new Date()).getTime());
    cmd = this.calculateWriteData(cmd);
    console.log('UART', 'Recursive write', cmd);
    // console.log('UART', noOfRegs, noOfDevices, getReadCmdTimeout(noOfRegs, noOfDevices));

    var opPromise = Q.defer();
    this.opPromise = opPromise;

    opPromise.promise.then(
      function() {
        if (cmdArr.length > 0) {
          this.recursiveWrite(u2a, cmdArr, deferred);
        } else {
          deferred.resolve();
        }
      }.bind(this)
    ).fail(function (err) {
      deferred.promise.reject(err);
    });

    u2a.sendCommandPacket(149, cmd);
  };

  CustomUART_Interface.prototype.manageWrites = function(u2a, cmdArr) {
    var deferred = Q.defer();
    if (cmdArr.length > 0) {
      this.recursiveWrite(u2a, cmdArr, deferred);
    } else {
      deferred.resolve();
    }
    return deferred.promise;
  };

  CustomUART_Interface.prototype.manageReadWrite = function(u2a, cmdInfo) {
    manageReadWriteUtil.bind(this)(u2a, cmdInfo);
  };

  /**
   * Construct and send a command packet for reading from a register
   *
   * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
   * @param info - The register map info for the register
   * @param regModel - the register model that is requesting the read
   * @param coreIndex - the core to read (undefined if not a deviceArray)
   * @return - A promise that is resolved once the data is received
   */
  CustomUART_Interface.prototype.read = function(
    u2a,
    info,
    regModel,
    coreIndex
  ) {
    console.log(
      'UART',
      'Read',
      u2a,
      info,
      regModel,
      coreIndex,
      getDeviceIndexUart(coreIndex, regModel._id),
      regReadDetail
    );
    var regAddr = info.addr;
    var readData = [],
      crc;

    var deviceIndex = getDeviceIndexUart(coreIndex, regModel._id);

    var regPage = document.querySelector('ti-widget-register-page');

    if (regReadDetail) {
      // if(regReadDetail && ((regReadDetail.regModelId === regModel._id) && ((coreIndex === undefined) || (regReadDetail.coreIndex === coreIndex)))) {

      if (deviceIndex < 0) {
        var deferred = Q.defer();

        regReadPromiseArr.push({
          promise: deferred,
          addr: info.addr
        });

        regReadCount++;

        if (regReadCount === regReadDetail.numRegs) {
          this.manageReadWrite(u2a, {
            type:
              this.is600 === true
                ? regModel._id === 'bq79600'
                  ? 'readAllBroadcast600'
                  : 'readAllStack'
                : 'readAllBroadcast',
            device: deviceIndex,
            count: regReadCount,
            promiseArr: regReadPromiseArr
          });
        }

        return deferred.promise;
      } else {
        var deferred = Q.defer();

        regReadPromiseArr.push({
          promise: deferred,
          addr: info.addr
        });

        regReadCount++;

        if (regReadCount === regReadDetail.numRegs) {
          this.manageReadWrite(u2a, {
            type:
              regModel._id === 'bq79600' ? 'readAllSingle600' : 'readAllSingle',
            device: deviceIndex,
            count: regReadCount,
            promiseArr: regReadPromiseArr
          });
        }

        return deferred.promise;
      }
    } else {
      regPage.isUpdating = true;
      console.log('UART', 'Read', 'Regmap - Show progress', regPage);

      if (deviceIndex < 0) {
        readData.push(0xc0); // Broadcast read

        regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
        readData.push.apply(readData, regAddr); // Register address
        readData.push(0x00); // Read 1 byte of data
        crc = getByteArrFromHexStr(
          prependZero(getCRC16IBM(readData).toString(16), 4)
        );
        readData.push.apply(readData, crc);

        var deferred = Q.defer();

        var readPromise = Q.defer();
        readPromise.promise.then(function(data) {
          var readData = [];
          data = parseDeviceResponse(data);
          for (var i = 0; i < data.length; i++) {
            readData.push(data[i][4]);
          }
          readData = readData.reverse();
          deferred.resolve(readData);
          regPage.isUpdating = false;
          console.log(
            'UART',
            'Read',
            'Broadcast read',
            'Regmap - Hide progress',
            regPage
          );
        }).fail(function (err) {
          deferred.promise.reject(err);
        });
        this.manageReadWrite(u2a, {
          type: 'read',
          data: readData,
          promise: readPromise
        });

        return deferred.promise;
      } else {
        readData.push(0x80); // Single device read intialization byte
        readData.push(deviceIndex); // Device address
        regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
        readData.push.apply(readData, regAddr); // Register address
        readData.push(0x00); // Read 1 byte of data
        crc = getByteArrFromHexStr(
          prependZero(getCRC16IBM(readData).toString(16), 4)
        );
        readData.push.apply(readData, crc); // CRC bytes

        var deferred = Q.defer();

        var readPromise = Q.defer();
        readPromise.promise.then(function(data) {
          deferred.resolve(data[4]);
          regPage.isUpdating = false;
          console.log(
            'UART',
            'Read',
            'Single read',
            'Regmap - Hide progress',
            regPage
          );
        }).fail(function (err) {
          deferred.promise.reject(err);
        });
        this.manageReadWrite(u2a, {
          type: 'read',
          data: readData,
          promise: readPromise
        });

        return deferred.promise;
      }
    }
  };

  /**
   * Construct and send a command packet for writing to a register
   *
   * @param u2a - Reference to the gc.databind.USB2ANY.prototype object
   * @param info - The register map info for the register
   * @param value - The value to write to the register
   * @param regModel - the register model that is requesting the write
   * @param coreIndex - the core to write (undefined if not a deviceArray)*
   * @return - A promise that resolves immediately
   */
  CustomUART_Interface.prototype.write = function(
    u2a,
    info,
    value,
    regModel,
    coreIndex
  ) {
    console.log('UART', 'Write', u2a, info, value, regModel, coreIndex);
    var writeData = [];
    var regAddr, crc;

    var deviceIndex = getDeviceIndexUart(coreIndex, regModel._id);

    if (deviceIndex > -1) {
      writeData.push(0x90); // Single device write intialization byte
      writeData.push(deviceIndex); // Device address
      regAddr = getByteArrFromHexStr(prependZero(info.addr.toString(16), 4));
      writeData.push.apply(writeData, regAddr); // Register address
      writeData.push(value); // Data byte
      crc = getByteArrFromHexStr(
        prependZero(getCRC16IBM(writeData).toString(16), 4)
      );
      writeData.push.apply(writeData, crc); // CRC bytes

      console.log('UART', 'Registermap', 'Write', writeData);
      this.manageReadWrite(u2a, {type: 'write', data: writeData});
    }

    return Q(true);
  };

  var logger = function(type, cmd) {
    document.querySelector('ti-bms-global').autoLogger(type, cmd);
  };

  var readRegsHdlr = function(eventType, event) {
    console.log('UART', 'Reg read event', eventType, event, event.detail);

    var detailObj = event.detail;
    if (detailObj.type === undefined) {
      detailObj.type = eventType;
    }

    var modelId = undefined;
    if (detailObj.regModelId) {
      modelId = detailObj.regModelId;
    }

    var deviceIndex = getDeviceIndexUart(detailObj.coreIndex, modelId);

    console.log(
      'bc: read_regs: numRegs=' +
        detailObj.numRegs +
        ', regModel id=' +
        modelId +
        ', coreIndex=' +
        detailObj.coreIndex +
        ', type=' +
        detailObj.type
    );
    if (
      detailObj.type === 'init' ||
      detailObj.type === 'button' ||
      detailObj.type === 'refresh'
    ) {
      // setTimeout(function() {
      // var modelId = "undefined";

      var regPage = document.querySelector('ti-widget-register-page');
      regPage.isUpdating = true;
      console.log('UART', 'Reg read event', 'Regmap - Show progress', regPage);

      if (regReadDetail && regReadDetail.modelId) {
        modelId = regReadDetail.modelId;
      }
      console.log(
        'regReadDetail.regModelId changed from ' +
          modelId +
          ' to ' +
          detailObj.regModelId
      );
      regReadDetail = detailObj;
      regReadDetail.device = deviceIndex;
      regReadCount = 0;
      regReadPromiseArr = [];
      // },0);
    }
  };

  var refreshPage = function() {
    var regPage = document.querySelector('ti-widget-register-page');
    regPage.readAllRegistersClickHdlr();
    templateObj.timerHdlr = window.setTimeout(
      refreshPage,
      templateObj.refreshInMs
    );
  };

  var templateObj = document.querySelector('#template_obj');
  templateObj.timerHdlr = null;
  templateObj.refreshInMs = 100;

  // templateObj.addEventListener('dom-change', function(event) {
  if (document.querySelector('ti-bms-global')) {
    document.addEventListener(
      'refresh-regs',
      readRegsHdlr.bind(templateObj, 'refresh')
    );
    document.addEventListener(
      'init-regs',
      readRegsHdlr.bind(templateObj, 'init')
    );
    document.addEventListener(
      'read_regs',
      readRegsHdlr.bind(templateObj, 'statusOrButton')
    );
    document.addEventListener('read_reg', function(event) {
      regReadDetail = undefined;
    });
    document.addEventListener('auto_read_config_changed', function(event) {
      if (templateObj.timerHdlr) {
        window.clearTimeout(templateObj.timerHdlr);
        templateObj.timerHdlr = null;
      }
      var autoReadObj = event.detail.autoRead;
      if (autoReadObj.value > -1) {
        templateObj.refreshInMs = +autoReadObj.value;
        templateObj.timerHdlr = window.setTimeout(
          refreshPage,
          templateObj.refreshInMs
        );
      }
    });
  }
  // });
};

templateObj = document.querySelector('#template_obj');
if (templateObj) {
  init();
} else {
  document.addEventListener('WebComponentsReady', init);
}
